import os
import h5py

def mkdir(path):
    if not os.path.exists(path):
        os.makedirs(path)

class generate_from_hdf5:
    def __init__(self, file, name):
        self.file_ = file
        self.name_ = name

    def __call__(self):
        with h5py.File(self.file_, 'r') as hf:
            for d in hf[self.name_]['table']:#d[0] is index name and d[1] is data
                yield d[1]

class runningMean(object):
    def __init__(self):
        self.n = 0
        self.mean = 0

    def append(self, newMean, size):
        self.mean = (self.mean * self.n + newMean * size)/(self.n + size)
        self.n += size



import tensorflow as tf
#tf.compat.v1.disable_v2_behavior()#This is for compatibility of shap and tf, might be able to remove in future
from CIMLA.ML.BaseModels import baseML
from CIMLA.ML.archs.MLP import build_MLP
from sklearn.pipeline import Pipeline
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import GridSearchCV
import joblib
import numpy as np
import xgboost as xgb
import copy


class RF(baseML):
    def __init__(self,
                 n_estimators,
                 max_depth,
                 max_features,
                 min_samples_leaf,
                 max_leaf_nodes,
                 cv,
                 scoring = 'accuracy',
                 n_jobs = 1):
        super(RF,self).__init__()

        pip = Pipeline(steps = [('rf', RandomForestClassifier(n_jobs = n_jobs, class_weight = 'balanced'))])
        params = {'rf__n_estimators': n_estimators,
                  'rf__max_depth': max_depth,
                  'rf__max_features': max_features,
                  'rf__min_samples_leaf': min_samples_leaf,
                  'rf__max_leaf_nodes': max_leaf_nodes
                  }

        self.gscv_ = GridSearchCV(pip, param_grid = params, cv = cv, refit = False, scoring = scoring)
        self.n_jobs_ = n_jobs
        self.model = None

    def _CV(self, sampled_date):
        """
        sampled_date is must be pandas df
        """
        x = sampled_date.iloc[:,:-1].values
        y = sampled_date.iloc[:,-1].values.reshape(-1)
        self.gscv_.fit(x, y)
        pars = self.gscv_.best_params_
        params = {'n_estimators': pars['rf__n_estimators'],
                  'max_depth': pars['rf__max_depth'],
                  'max_features': pars['rf__max_features'],
                  'min_samples_leaf': pars['rf__min_samples_leaf'],
                  'max_leaf_nodes': pars['rf__max_leaf_nodes']}

        print(params)

        return params

    def _build_model(self, pars, nSample, cache = None, batch_size = 3000):
        if cache:
            warm_start = True
            n_batch = np.ceil(nSample/batch_size)#this is an approximate, but best we can do!
            n_estimators = int(max(1, pars['n_estimators']//n_batch))
        else:
            warm_start = False
            n_estimators = pars['n_estimators']

        model = RandomForestClassifier(n_estimators = n_estimators,
                                      max_depth = pars['max_depth'],
                                      max_features = pars['max_features'],
                                      min_samples_leaf = pars['min_samples_leaf'],
                                      max_leaf_nodes = pars['max_leaf_nodes'],
                                      n_jobs = self.n_jobs_,
                                      warm_start = warm_start,
                                      class_weight = 'balanced')
        return model, n_estimators

    def _batch_training(self, data, model, n_estimators):
        for b in data.batch(split = 'train', size = 3000, shuffle = True):
            curr = b.numpy()
            model.fit(X = curr[:,:-1], y = curr[:,-1].reshape(-1,))
            model.n_estimators += n_estimators
        model.n_estimators -= n_estimators
        return model

    def train(self, data):
        # first do cv on <=3000 samples to select best parameters
        sampled_date = data.sample(n = 6000, split = 'train')
        if data.cache_: #data is dask dataframe, should convert it to pandas
            sampled_date = sampled_date.compute()
        best_pars = self._CV(sampled_date)

        # train the final model
        model, n_estimators = self._build_model(pars = best_pars, nSample = data.nSample['train'], cache = data.cache_, batch_size = 6000)
        if data.cache_: #batch training is needed
            self.model = self._batch_training(data, model, n_estimators)
        else: #normal training, data is pandas dataframe
            train_data = data.getDatadf(split = 'train')
            model.fit(X = train_data.iloc[:,:-1], y = train_data.iloc[:,-1])
            self.model = model

    def predict(self, x):
        if not self.model:
            raise ValueError('model has not yet been trained.')
        return self.model.predict(x)

    def save(self, path):
        if not self.model:
            raise ValueError('model has not yet been trained.')
        joblib.dump(self.model, path)

class deepMLP(baseML):
    def __init__(self, input_shape, mid_channels, l2 = 0, dropout = None, class_weight = None):
        super(deepMLP,self).__init__()
        self.model = build_MLP(input_shape, mid_channels, l2 = l2, dropout = dropout)
        self.model.compile(optimizer = "Adam", loss = tf.keras.losses.BinaryCrossentropy(from_logits = True))
        self.classW_ = class_weight

    def _tarin_on_batch(self, batch_x, batch_y):
        self.model.train_on_batch(x = batch_x, y = batch_y, reset_metrics = False, class_weight = self.classW_)

    def train(self, data, batch_size = 128, epochs = 100, shuffle = True):
        if data.cache_:
            batches = data.batch(split = 'train', size = batch_size, shuffle = shuffle)
            for _ in range(epochs):
                for b in batches:
                    self._tarin_on_batch(b[:,:-1], b[:,-1])
        else:
            df = data.getDatadf(split = 'train')
            X = df.values[:,:-1]
            y = df.values[:,-1]
            self.model.fit(x = X, y = y, batch_size = batch_size, epochs = epochs, shuffle = shuffle, verbose = 0, class_weight = self.classW_)

    def predict(self, x, tresh = 0.5):
        logit = self.model.predict_on_batch(x = x)
        p = np.exp(logit) / (1 + np.exp(logit))
        #p = self.model.predict_on_batch(x = x)

        if tresh:
            return p >= tresh
        else:
            return p

    def save(self, path):
        #TODO
        pass

####TOdos
# if possible train with logit but run shap on prob, or modify code ...
#



class baseML(object):
    def __init__(self):
        self.model = None
        pass

    def train(self, data):
        pass

    def predict(self, x):
        pass

    def save(self, path):
        pass



from CIMLA.attribution.BaseModels import baseAM
import shap
import numpy as np

class treeSHAP(baseAM):
    def __init__(self):
        super(treeSHAP,self).__init__()
        self.local_feature_importance = None

    def set_explainer(self, ML_model, background = None, model_output = "raw"):
        self.explainer_ = shap.TreeExplainer(ML_model.model, model_output = model_output)

    def explain(self, x):
        return self.explainer_.shap_values(x)

class deepSHAP(baseAM):
    def __init__(self):
        super(deepSHAP,self).__init__()
        self.local_feature_importance = None

    def set_explainer(self, ML_model, background, model_output = "raw"):
        self.explainer_ = shap.DeepExplainer(ML_model.model, background)

    def explain(self, x):
        return self.explainer_.shap_values(x)[0]



import yaml
import CIMLA.defaults as defaults
import pandas as pd
import numpy as np
import dask.dataframe as dd
from CIMLA.Dataset import data
from CIMLA.ML.Regression import xgboost
#from ML.Classification import RF_clf
from CIMLA.attribution.SHAP import treeSHAP, deepSHAP
from CIMLA.CIMLA import cimla
from CIMLA.PostProcessor import processor

"""
TODOs:
1. write a checker method that makes sure the provided flags are legit
"""

class parser(object):

    @staticmethod
    def parse(config_file):
        with open(config_file,'r') as cf:
            config = yaml.safe_load(cf)

        config = defaults.params(config)
        data_g1, data_g2, n_features,classW1, classW2 = parser._parse_data(config['data'], task = config['ML']['task'])
        ML_g1, ML_g2 = parser._parse_ML(config['ML'], n_features, class_weights = (classW1, classW2))
        att_g1, att_g2 = parser._parse_attribution(config['attribution'])
        pipeline = cimla(data_g1 = data_g1,
                         data_g2 = data_g2,
                         ML_g1 = ML_g1,
                         ML_g2 = ML_g2,
                         attr_g1 = att_g1,
                         attr_g2 = att_g2)
        proc = parser._parse_postproc(config['post_process'], pipeline)
        return pipeline, proc, config

    @staticmethod
    def _to_classification(df1,df2,y_name, type='median'):

        t = df1[y_name].values.flatten().tolist() + df2[y_name].values.flatten().tolist()
        t = np.median(t)

        new = []
        for _,r in df1.iterrows():
            if r[y_name] > t:
                new.append(1)
            else:
                new.append(0)
        df1[y_name] = new
        pos = sum(new)
        neg = len(new) - pos
        total = len(new)
        weights_1 = {0: (1 / neg) * (total / 2.0), 1: (1 / pos) * (total / 2.0)}

        new = []
        for _,r in df2.iterrows():
            if r[y_name] > t:
                new.append(1)
            else:
                new.append(0)
        df2[y_name] = new
        pos = sum(new)
        neg = len(new) - pos
        total = len(new)
        weights_2 = {0: (1 / neg) * (total / 2.0), 1: (1 / pos) * (total / 2.0)}

        return df1,df2, weights_1, weights_2

    @staticmethod
    def _parse_data(data_config, task = 'regression'):
        if data_config['cache']:#use dask
            index_col = data_config['sample_column'] if data_config['sample_column'] else 'Unnamed: 0'
            df1 = dd.read_csv(data_config['path_g1'], header = 0, dtype = {index_col : 'string'})#.set_index(index_col)
            df2 = dd.read_csv(data_config['path_g2'], header = 0, dtype = {index_col : 'string'})#.set_index(index_col)
        else:
            df1 = pd.read_csv(data_config['path_g1'], header = 0, index_col = 0)
            df2 = pd.read_csv(data_config['path_g2'], header = 0, index_col = 0)


        df1.columns = df1.columns.astype(str)
        df2.columns = df2.columns.astype(str)

        indep = pd.read_csv(data_config['independent_parameters'], header = None, index_col = None, dtype = 'string').values.flatten().tolist()
        dep = pd.read_csv(data_config['dependent_parameters'], header = None, index_col = None, dtype = 'string').values.flatten().tolist()

        weights_1 = weights_2 = None
        if task == 'classification':
            df1,df2,weights_1, weights_2 = parser._to_classification(df1, df2, dep[0], type='median')


        data_g1 = data(data = df1,
                       nSample = data_config['number_samples_g1'],
                       indep_params = indep.copy(),
                       dep_params = dep,
                       normalize = data_config['normalize'],
                       test_size = data_config['test_size'],
                       split_rnd_state = data_config['split_random_state'],
                       cache = data_config['cache'],
                       task = task,
                       label = 'g1')

        data_g2 = data(data = df2,
                       nSample = data_config['number_samples_g2'],
                       indep_params = indep.copy(),
                       dep_params = dep,
                       normalize = data_config['normalize'],
                       test_size = data_config['test_size'],
                       split_rnd_state = data_config['split_random_state'],
                       cache = data_config['cache'],
                       task = task,
                       label = 'g2')

        return data_g1, data_g2, len(indep), weights_1, weights_2

    @staticmethod
    def _parse_ML(ML_config, n_features, class_weights = None):
        if ML_config['type'] == 'RF':
            if ML_config['task'] == 'regression':
                from ML.Regression import RF
            else:
                from ML.Classification import RF

            max_depth = [i for i in ML_config['max_depth'] if i != 'None']
            if 'None' in ML_config['max_depth']:
                max_depth.append(None)

            max_features = [i for i in ML_config['max_features'] if i != 'None']
            if 'None' in ML_config['max_features']:
                max_features.append(None)


            model1 = RF(n_estimators = ML_config['n_estimators'],
                        max_depth = ML_config['max_depth'],
                        max_features = ML_config['max_features'],
                        min_samples_leaf = ML_config['min_samples_leaf'],
                        max_leaf_nodes = ML_config['max_leaf_nodes'],
                        cv = ML_config['cv'],
                        scoring = ML_config['scoring'],
                        n_jobs = 1)

            model2 = RF(n_estimators = ML_config['n_estimators'],
                        max_depth = ML_config['max_depth'],
                        max_features = ML_config['max_features'],
                        min_samples_leaf = ML_config['min_samples_leaf'],
                        max_leaf_nodes = ML_config['max_leaf_nodes'],
                        cv = ML_config['cv'],
                        scoring = ML_config['scoring'],
                        n_jobs = 1)

        elif ML_config['type'] == 'MLP':
            if ML_config['task'] == 'regression':
                from ML.Regression import deepMLP
            else:
                from ML.Classification import deepMLP

            model1 = deepMLP(input_shape = n_features, mid_channels = ML_config['hidden_channels'], l2 = ML_config['dense_layers_l2'], dropout = ML_config['dropout'], class_weight = class_weights[0])
            model2 = deepMLP(input_shape = n_features, mid_channels = ML_config['hidden_channels'], l2 = ML_config['dense_layers_l2'],dropout = ML_config['dropout'], class_weight = class_weights[1])
        elif ML_config['type'] == 'XGB':
            model1 = xgboost(max_depth = ML_config['max_depth'],
                             min_child_weight = ML_config['min_child_weight'],
                             subsample = ML_config['subsample'],
                             colsample_bynode = ML_config['colsample'],
                             eta = ML_config['eta'],
                             l2 = ML_config['l2'],
                             l1 = ML_config['l1'],
                             max_boost_round = ML_config['max_boost_round'],
                             early_stopping_round = ML_config['early_stop_round'],
                             early_stopping_tol = ML_config['early_stop_tol'],
                             scoring = ML_config['scoring'],
                             objective = 'reg:squarederror')

            model2 = xgboost(max_depth = ML_config['max_depth'],
                             min_child_weight = ML_config['min_child_weight'],
                             subsample = ML_config['subsample'],
                             colsample_bynode = ML_config['colsample'],
                             eta = ML_config['eta'],
                             l2 = ML_config['l2'],
                             l1 = ML_config['l1'],
                             max_boost_round = ML_config['max_boost_round'],
                             early_stopping_round = ML_config['early_stop_round'],
                             early_stopping_tol = ML_config['early_stop_tol'],
                             scoring = ML_config['scoring'],
                             objective = 'reg:squarederror')
        else:
            raise ValueError('{} machine learning model is not yet supported.'.format(ML_config['type']))

        return model1, model2

    @staticmethod
    def _parse_attribution(att_config):
        if att_config['type'] == 'tree_shap':
            model1 = treeSHAP()
            model2 = treeSHAP()
        elif att_config['type'] == 'deep_shap':
            model1 = deepSHAP()
            model2 = deepSHAP()
        else:
            raise ValueError('"{}" attribution model is not supported.'.format(att_config['type']))

        return model1, model2

    @staticmethod
    def _parse_postproc(proc_config, pipeline):
        proc = processor(CIMLA = pipeline,
                         save_local = proc_config['local_scores_path'],
                         save_global = proc_config['global_scores_path'],
                         save_ML = proc_config['ML_save_path'],
                         save_performance = proc_config['ML_performance_save_path'],
                         metric = proc_config['ML_performance_metric'])
        return proc



"""
TODOs:
1. organize key default params
"""
default_params = {'batch':-1,
                  'shuffle':False,}

"""
functions
"""

def params(config):
    for k in default_params.keys():
        if k not in config.keys():
            config[k] = default_params[k]

    return config



import pandas as pd
import numpy as np
import sklearn.preprocessing
import dask_ml.preprocessing
import sklearn.model_selection
import dask_ml.model_selection
import os
os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
os.environ['TF_CPP_MIN_LOG_LEVEL'] = "3"  #only display errors
import tensorflow as tf
import h5py
from CIMLA.utils import mkdir, generate_from_hdf5
#tf.compat.v1.disable_v2_behavior()#This is for compatibility of shap and tf, might be able to remove in future

class data(object):
    def __init__(self,
                 data,
                 nSample,
                 indep_params,
                 dep_params,
                 normalize,
                 test_size,
                 split_rnd_state = None,
                 cache = None,
                 task = 'regression',
                 label = 'g1'):

        assert(isinstance(indep_params,list))
        assert(isinstance(dep_params,list))
        #For now only support single dependent parameter
        assert(len(dep_params) == 1)
        self.task = task
        self.label_ = label
        self.cache_ = cache
        self.indep_pars_ = indep_params
        self.dep_pars_ = dep_params
        df = self._sort_xy(data, nSample, normalize, task)
        self.data_ = self._split(df = df, test_size = test_size, random_state = split_rnd_state)
        self.nSample = {'train': nSample * (1-test_size), 'test': nSample * test_size}
        if cache:
            mkdir(cache)
            self.generators_ = {}
            self.generators_['train'] = self._writeHdf5(cache, self.data_['train'], 'train')
            self.generators_['test'] = self._writeHdf5(cache, self.data_['test'], 'test')

    def _sort_xy(self, data_df, nSample, normalize = True, task = 'regression'):
        """
        sets last column to be Y
        """
        if task == 'classification':
            df = data_df[self.indep_pars_]
            dep_par_name = 'dependent_{}'.format(self.dep_pars_[0])
            cols = df.columns
        else:
            df = data_df[self.indep_pars_ + self.dep_pars_]
            cols = df.columns[:-1].tolist() + ['dependent_{}'.format(df.columns[-1])]
            df.columns = cols

        if normalize:
            scaler = sklearn.preprocessing.StandardScaler() if not self.cache_ else dask_ml.preprocessing.StandardScaler()
            df = scaler.fit_transform(df)

        if not self.cache_:#should transfer back to pandas DataFrame
            df = pd.DataFrame(df, columns = cols, index = data_df.index)

        # randomize/permute the independent parameter if it's the same as the target parameter
        for p in self.dep_pars_:
            if p in self.indep_pars_:
                if self.cache_:
                    import dask.array as da
                    import dask.dataframe as dd
                    tmp = da.random.normal(size = nSample, chunks = tuple(df.map_partitions(len).compute()))
                    tmp = dd.from_dask_array(tmp, columns=['rand'], index = df.index)
                    df[p] = tmp['rand']
                else:
                    perm = np.random.permutation(df[p].values.flatten())
                    df[p] = perm

        if task == 'classification':
            if self.cache_:
                import dask.array as da
                import dask.dataframe as dd
                tmp = dd.from_array(data_df[self.dep_pars_].values.flatten().reshape(-1,1), columns=['y'], index = df.index)
                df[self.dep_pars_] = tmp['y']
            else:
                df.insert(df.shape[-1], dep_par_name, data_df[self.dep_pars_])

        return df


    def _split(self, df, test_size, random_state = None):
        data = {}
        if not self.cache_:
            split = sklearn.model_selection.train_test_split
            data['train'], data['test'] = split(df, random_state = random_state, test_size = test_size)
        else:
            data['train'], data['test'] = df.random_split([1-test_size, test_size], random_state = random_state)
        return data

    def sample(self, n, split = 'train'):
        """
        n should be an integer, a data with approximately n samples is returned
        We used frac argument since "n" is not supported in dask dataframe
        """
        frac = min(1, np.round(n/self.nSample[split],4))
        return self.data_[split].sample(frac = frac, replace = False)

    #def _writeHdf5(self, path, data, name):
    #    with h5py.File(path+'/data_{}.hdf5'.format(self.label_), 'a') as f:
    #        f.create_dataset(name = name, data = data, dtype = 'float32')

    #    generator = generate_from_hdf5(path+'/data_{}.hdf5'.format(self.label_), name)
    #    return generator

    def _writeHdf5(self, path, data, name):
        data.to_hdf(path+'/data_{}.hdf5'.format(self.label_), key = name)
        generator = generate_from_hdf5(path+'/data_{}.hdf5'.format(self.label_), name)
        return generator

    def _append_to_generators(self, data, prefix = 'sample'):
        id = len(self.generators_.keys()) - 2 #there are two generators by default ('train' and 'test')
        name = '{}_{}'.format(prefix,id)
        self.generators_[name] = self._writeHdf5(self.cache_, data, name)
        return name

    def batch(self, split, size, shuffle = True, sample_n = None):
        if not self.cache_:
            raise ValueError('batch creation is only supported when an hdf5 cache folder is specified.')

        if sample_n:
            sampled_date = self.sample(sample_n, split)
            split = self._append_to_generators(sampled_date)

        nPar = len(self.indep_pars_) + 1
        ds = tf.data.Dataset.from_generator(self.generators_[split], tf.float32, tf.TensorShape([nPar]))
        if shuffle:
            ds = ds.shuffle(buffer_size = 1000, reshuffle_each_iteration = True)

        return ds.batch(size, drop_remainder = False)

    def getDatadf(self, split = 'train'):
        """
        returns dask dataframe if self.cache_ is specified, otherwise returns pandas dataframe
        """
        if split not in ['train','test']:
            raise ValueError('specified {} split is not available.'.format(split))

        return self.data_[split]



import numpy as np
class baseAM(object):
    def __init__(self):
        self.explainer_ = None
        self.local_feature_importance = None

    def set_explainer(self, ML_model, background = None):
        """
        ML_model: is an object of type baseML model
        """
        pass

    def explain(self, x):
        """
        x is a numpy array
        """
        pass

    def append_local_importance(self, importance):
        if self.local_feature_importance:
            self.local_feature_importance = np.append(self.local_feature_importance, importance, axis = 0)
        else:
            self.local_feature_importance = importance



import pandas as pd
import numpy as np
from CIMLA.attribution.metrics import RMSD
import joblib
from CIMLA.utils import mkdir, runningMean
from sklearn.metrics import r2_score, accuracy_score, mean_squared_error
import shutil


class processor(object):
    def __init__(self,
                 CIMLA,
                 save_local = None,
                 save_global = None,
                 save_ML = None,
                 save_performance = None,
                 metric = 'mse'):


        self.cimla_ = CIMLA
        self.local_path_ = save_local
        self.global_path_ = save_global
        self.ml_path_ = save_ML
        self.perf_path_ = save_performance
        self.metric_ = metric

    def _save_local(self, path):
        g1 = self.cimla_.attr_['g1'].local_feature_importance
        g2 = self.cimla_.attr_['g2'].local_feature_importance
        g1 = pd.DataFrame(g1)
        g2 = pd.DataFrame(g2)
        g1.columns = self.cimla_.data_['g1'].indep_pars_
        g2.columns = self.cimla_.data_['g2'].indep_pars_
        g1.to_csv(path + '/local_feature_importance_g1.csv', header = True, index = True)
        g2.to_csv(path + '/local_feature_importance_g2.csv', header = True, index = True)


    def _save_global(self,path):
        scores = self.cimla_.feature_importance.reshape(1,-1)
        scores = pd.DataFrame(scores)
        scores.columns = self.cimla_.data_['g1'].indep_pars_
        scores.to_csv(path + '/global_feature_importance.csv', header = True, index = True)


    def _save_ml(self, path):
        ml1 = self.cimla_.ml_['g1'].model
        ml2 = self.cimla_.ml_['g2'].model
        joblib.dump(ml1, path + '/ML_g1')
        joblib.dump(ml2, path + '/ML_g2')

    def _compute_perf_on_group_and_split(self, g, split, function):
        if self.cimla_.data_[g].cache_: #use batch
            perf = runningMean()
            batches = self.cimla_.data_[g].batch(split = split, size = 128, shuffle = False, sample_n = None)
            y_pred = []
            y_true = []
            for b in batches:
                currB = b.numpy()
                X = currB[:,:-1]
                y = currB[:,-1]
                y_hat = self.cimla_.ml_[g].predict(X)
                currPerf = function(y_hat, y)
                currSize = X.shape[0]
                perf.append(currPerf, currSize)
                y_pred.append(y_hat.flatten())
                y_true.append(y.flatten())
            return perf.mean, np.concatenate(y_pred).reshape(-1,1), np.concatenate(y_true).reshape(-1,1)
            ## TODO: make sure the returned y_pred has the right shape

        else:
            data = self.cimla_.data_[g].getDatadf(split = split).values
            X = data[:,:-1]
            y = data[:,-1]
            y_hat = self.cimla_.ml_[g].predict(X)
            return function(y, y_hat),y_hat.reshape(-1,1), y.reshape(-1,1)


    def _compute_perf(self, function):
        train_g1, yP_g1_train, yT_g1_train = self._compute_perf_on_group_and_split(g = 'g1', split = 'train', function = function)
        train_g2, yP_g2_train, yT_g2_train = self._compute_perf_on_group_and_split(g = 'g2', split = 'train', function = function)
        test_g1, yP_g1_test, yT_g1_test = self._compute_perf_on_group_and_split(g = 'g1', split = 'test', function = function)
        test_g2, yP_g2_test, yT_g2_test = self._compute_perf_on_group_and_split(g = 'g2', split = 'test', function = function)

        perf = np.array([train_g1, test_g1, train_g2, test_g2]).reshape(1,-1)
        perf = pd.DataFrame(perf)
        perf.columns = ['train_g1', 'test_g1', 'train_g2', 'test_g2']

        y_g1_train = pd.DataFrame(np.concatenate([yT_g1_train, yP_g1_train], axis = 1))
        y_g2_train = pd.DataFrame(np.concatenate([yT_g2_train, yP_g2_train], axis = 1))
        y_g1_test = pd.DataFrame(np.concatenate([yT_g1_test, yP_g1_test], axis = 1))
        y_g2_test = pd.DataFrame(np.concatenate([yT_g2_test, yP_g2_test], axis = 1))

        y_g1_train.columns = ['y_true','y_pred']
        y_g2_train.columns = ['y_true','y_pred']
        y_g1_test.columns = ['y_true','y_pred']
        y_g2_test.columns = ['y_true','y_pred']

        return perf, y_g1_train, y_g2_train, y_g1_test, y_g2_test

    def _save_perf(self, path):
        if self.metric_ == 'r2':
            func = r2_score
        elif self.metric_ == 'accuracy':
            func = accuracy_score
        elif self.metric_ == 'mse':
            func = mean_squared_error
        else:
            raise ValueError('ML evaluation metric {} is not implemented.'.format(self.metric_))

        perf, y_g1_train, y_g2_train, y_g1_test, y_g2_test = self._compute_perf(func)
        perf.to_csv(path + '/ML_performance.csv', header = True, index = True)

        y_g1_train.to_csv(path + '/y_g1_train.csv', header = True, index = True)
        y_g2_train.to_csv(path + '/y_g2_train.csv', header = True, index = True)
        y_g1_test.to_csv(path + '/y_g1_test.csv', header = True, index = True)
        y_g2_test.to_csv(path + '/y_g2_test.csv', header = True, index = True)

    def process(self):
        if self.local_path_:
            mkdir(self.local_path_)
            self._save_local(self.local_path_)
        if self.global_path_:
            mkdir(self.global_path_)
            self._save_global(self.global_path_)
        if self.ml_path_:
            mkdir(self.ml_path_)
            self._save_ml(self.ml_path_)
        if self.perf_path_:
            mkdir(self.perf_path_)
            self._save_perf(self.perf_path_)

    def clear(self):
        if self.cimla_.data_['g1'].cache_:
            shutil.rmtree(self.cimla_.data_['g1'].cache_)



import numpy as np
from CIMLA.attribution.SHAP import deepSHAP, treeSHAP
from CIMLA.attribution.metrics import RMSD, MSD
from CIMLA.utils import runningMean
class cimla (object):
    def __init__(self,
                 data_g1,
                 data_g2,
                 ML_g1,
                 ML_g2,
                 attr_g1,
                 attr_g2):

        """
        attr_data: data split used for explanation: 'train', 'test'
        """

        self.data_ = {'g1': data_g1,
                      'g2': data_g2}

        self.ml_ = {'g1': ML_g1,
                    'g2': ML_g2}

        self.attr_ = {'g1': attr_g1,
                     'g2': attr_g2}

        self.feature_importance = None
        self.task = data_g1.task
        if self.task == 'regression':
            self.ml_out_ = 'raw'
        else:
            self.ml_out_ = 'raw'

    def run(self, attr_data_split = 'train', attr_data_group = 2, attr_data_size = 1, global_type = 'rmsd'):
        dg = 'g' + str(attr_data_group)
        sample_size = self.data_[dg].nSample[attr_data_split] * attr_data_size if attr_data_size < 1 else None
        attr_data = self._get_attr_data(data = self.data_[dg], split = attr_data_split, sample_size = sample_size)
        background_x = self._get_background_data(data = self.data_[dg], split = attr_data_split, backgroud_size = 1000)# if isinstance(self.attr_['g1'], deepSHAP) else None

        print("* Trainig ML Models ...")
        print("    -- training ML model for group 1 ...")
        self._trainML(self.ml_['g1'], self.data_['g1'])
        print("    -- training ML model for group 2 ...")
        self._trainML(self.ml_['g2'], self.data_['g2'])


        print("* Running Attribution Model ...")
        for g in ['g1','g2']:
            self.attr_[g].set_explainer(self.ml_[g], background_x, model_output = self.ml_out_)
            #print(self.attr_[g].explainer_.data.shape[0])
        self.feature_importance = self._runAM(attr_data, global_type)


    def _trainML(self, model, data):
        model.train(data)

    def _get_attr_data(self, data, split, sample_size = None):
        if data.cache_:
            return data.batch(split, 1024, shuffle = True, sample_n = sample_size)
        elif sample_size:
            return data.sample(n = sample_size, split = split).iloc[:,:-1].values
        else:
            return data.getDatadf(split).iloc[:,:-1].values


    def _get_background_data(self, data, split, backgroud_size):
        bg = data.sample(backgroud_size, split).iloc[:,:-1]
        if data.cache_:
            bg = bg.compute()
        return bg.values

    def _compute_global(self, type, feat_importance_g1, feat_importance_g2):
        if type == 'rmsd':
            func = RMSD
        elif type == 'msd':
            func = MSD
        else:
            raise ValueError('global attribution score {} is not implemented.'.format(type))

        return func(s1 = feat_importance_g1, s2 = feat_importance_g2)

    def _runAM(self, attr_data, global_type, store_local = False):
        if isinstance(attr_data, np.ndarray):
            g1 = self.attr_['g1'].explain(attr_data)
            g2 = self.attr_['g2'].explain(attr_data)
            if isinstance(g1,list):
                g1 = g1[-1] #contribution to the probability of belonging to the last group
                g2 = g2[-1] #contribution to the probability of belonging to the last group

            if store_local:
                self.attr_['g1'].append_local_importance(g1)
                self.attr_['g2'].append_local_importance(g2)
            return self._compute_global(global_type, g1, g2)

        else: #data is tensorflow dataset with batches
            score = runningMean()
            global_type = 'msd' if global_type == 'rmsd' else global_type
            for b in attr_data:
                currX = b.numpy()[:,:-1]
                g1 = self.attr_['g1'].explain(currX)
                g2 = self.attr_['g2'].explain(currX)
                if isinstance(g1,list):
                    g1 = g1[-1] #contribution to the probability of belonging to the last group
                    g2 = g2[-1] #contribution to the probability of belonging to the last group

                if store_local:
                    self.attr_['g1'].append_local_importance(g1)
                    self.attr_['g2'].append_local_importance(g2)

                curr = self._compute_global(global_type, g1, g2)
                size = currX.shape[0]
                score.append(curr, size)
                if global_type == 'msd':#return rmsd
                    ret = np.power(score.mean, 0.5)
            return ret



import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

VERSION = '0.0.13'
PACKAGE_NAME = 'CIMLA'
AUTHOR = 'Payam Dibaeinia'
AUTHOR_EMAIL = 'dibaein2@illinois.edu'
URL = 'https://github.com/PayamDiba/CIMLA'

LICENSE = 'MIT License'
DESCRIPTION = 'Counterfactual Inference by Machine Learning and Attribution Models'
LONG_DESCRIPTION = (HERE / "README.md").read_text()
LONG_DESC_TYPE = "text/markdown"

INSTALL_REQUIRES = [
      'dask==2021.10.0',
      'h5py==2.10.0',
      'numpy==1.21.2',
      'pandas==1.3.3',
      'scipy',
      'tensorflow==2.2.0',
      'dask-ml==1.9.0',
      'tables==3.6.*',
      'shap==0.39.0',
      'scikit-learn==0.24.2',
      'xgboost==1.5.0',
      'PyYAML==5.4.1',
      'protobuf==3.20',
]

setup(name=PACKAGE_NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      long_description_content_type=LONG_DESC_TYPE,
      author=AUTHOR,
      license=LICENSE,
      author_email=AUTHOR_EMAIL,
      url=URL,
      install_requires=INSTALL_REQUIRES,
      packages=find_packages(),
      entry_points ={'console_scripts': [ 'cimla = CIMLA.main:main']},
      python_requires='==3.8.12',
      )



import argparse
from CIMLA.ConfigParser import parser
"""
TODOs:
1. Move post processor to CIMLA_pipeline --> helps making it available as a module
"""

def main():

    flag = argparse.ArgumentParser()
    flag.add_argument('--config', type = str, default = None, help = 'path to config file', required = True)
    config_file = flag.parse_args().config
    CIMLA_pipeline, postProcessor, confs = parser.parse(config_file)
    CIMLA_pipeline.run(attr_data_split = confs['attribution']['data_split'],
                       attr_data_group = confs['attribution']['data_group'],
                       attr_data_size = confs['attribution']['data_size'],
                       global_type = confs['attribution']['global_type'])
    print("* Post-processing ...")
    postProcessor.process()
    #postProcessor.clear()

if __name__ == "__main__":
    main()



import tensorflow as tf
#tf.compat.v1.disable_v2_behavior()#This is for compatibility of shap and tf, might be able to remove in future
from CIMLA.ML.BaseModels import baseML
from CIMLA.ML.archs.MLP import build_MLP
from sklearn.pipeline import Pipeline
from sklearn.ensemble import RandomForestRegressor
from sklearn.model_selection import GridSearchCV
import joblib
import numpy as np
import xgboost as xgb
import copy


class RF(baseML):
    def __init__(self,
                 n_estimators,
                 max_depth,
                 max_features,
                 min_samples_leaf,
                 max_leaf_nodes,
                 cv,
                 scoring = 'neg_mean_squared_error',
                 n_jobs = 1):
        super(RF,self).__init__()

        pip = Pipeline(steps = [('rf', RandomForestRegressor(n_jobs = n_jobs))])
        params = {'rf__n_estimators': n_estimators,
                  'rf__max_depth': max_depth,
                  'rf__max_features': max_features,
                  'rf__min_samples_leaf': min_samples_leaf,
                  'rf__max_leaf_nodes': max_leaf_nodes
                  }

        self.gscv_ = GridSearchCV(pip, param_grid = params, cv = cv, refit = False, scoring = scoring)
        self.n_jobs_ = n_jobs
        self.model = None

    def _CV(self, sampled_date):
        """
        sampled_date is must be pandas df
        """
        x = sampled_date.iloc[:,:-1].values
        y = sampled_date.iloc[:,-1].values.reshape(-1)
        self.gscv_.fit(x, y)
        pars = self.gscv_.best_params_
        params = {'n_estimators': pars['rf__n_estimators'],
                  'max_depth': pars['rf__max_depth'],
                  'max_features': pars['rf__max_features'],
                  'min_samples_leaf': pars['rf__min_samples_leaf'],
                  'max_leaf_nodes': pars['rf__max_leaf_nodes']}

        print("    -- final model: ", params)

        return params

    def _build_model(self, pars, nSample, cache = None, batch_size = 3000):
        if cache:
            warm_start = True
            n_batch = np.ceil(nSample/batch_size)#this is an approximate, but best we can do!
            n_estimators = int(max(1, pars['n_estimators']//n_batch))
        else:
            warm_start = False
            n_estimators = pars['n_estimators']

        model = RandomForestRegressor(n_estimators = n_estimators,
                                      max_depth = pars['max_depth'],
                                      max_features = pars['max_features'],
                                      min_samples_leaf = pars['min_samples_leaf'],
                                      max_leaf_nodes = pars['max_leaf_nodes'],
                                      n_jobs = self.n_jobs_,
                                      warm_start = warm_start)
        return model, n_estimators

    def _batch_training(self, data, model, n_estimators):
        for b in data.batch(split = 'train', size = 6000, shuffle = True):
            curr = b.numpy()
            model.fit(X = curr[:,:-1], y = curr[:,-1].reshape(-1,))
            model.n_estimators += n_estimators
        model.n_estimators -= n_estimators
        return model

    def train(self, data):
        # first do cv on <=6000 samples to select best parameters
        sampled_date = data.sample(n = 6000, split = 'train')
        if data.cache_: #data is dask dataframe, should convert it to pandas
            sampled_date = sampled_date.compute()
        best_pars = self._CV(sampled_date)

        # train the final model
        model, n_estimators = self._build_model(pars = best_pars, nSample = data.nSample['train'], cache = data.cache_, batch_size = 6000)
        if data.cache_: #batch training is needed
            self.model = self._batch_training(data, model, n_estimators)
        else: #normal training, data is pandas dataframe
            train_data = data.getDatadf(split = 'train')
            model.fit(X = train_data.iloc[:,:-1], y = train_data.iloc[:,-1])
            self.model = model

    def predict(self, x):
        if not self.model:
            raise ValueError('model has not yet been trained.')
        return self.model.predict(x)

    def save(self, path):
        if not self.model:
            raise ValueError('model has not yet been trained.')
        joblib.dump(self.model, path)




class deepMLP(baseML):
    def __init__(self, input_shape, mid_channels, l2 = 0, dropout = None, class_weight = None):
        super(deepMLP,self).__init__()
        self.model = build_MLP(input_shape, mid_channels, l2 = l2, dropout = dropout)
        self.model.compile(optimizer = "Adam", loss = "mse")
        self.classW_ = class_weight

    def _tarin_on_batch(self, batch_x, batch_y):
        self.model.train_on_batch(x = batch_x, y = batch_y, reset_metrics = False, class_weight = self.classW_)

    def train(self, data, batch_size = 128, epochs = 100, shuffle = True, class_weight = None):
        if data.cache_:
            batches = data.batch(split = 'train', size = batch_size, shuffle = shuffle)
            for _ in range(epochs):
                for b in batches:
                    self._tarin_on_batch(b[:,:-1], b[:,-1])
        else:
            df = data.getDatadf(split = 'train')
            X = df.values[:,:-1]
            y = df.values[:,-1]
            self.model.fit(x = X, y = y, batch_size = batch_size, epochs = epochs, shuffle = shuffle, verbose = 0, class_weight = self.classW_)

    def predict(self, x):
        return self.model.predict_on_batch(x = x)

    def save(self, path):
        #TODO
        pass


class xgboost(baseML):
    def __init__(self,
                 max_depth,
                 min_child_weight,
                 subsample,
                 colsample_bynode,
                 eta,
                 l2,
                 l1,
                 max_boost_round = 250,
                 early_stopping_round = 30,#must be specified ## TODO
                 early_stopping_tol = 1e-2,
                 scoring = 'mae',
                 objective = 'reg:squarederror'):
        super(xgboost,self).__init__()

        self.grid_param_ = {
            'max_depth':max_depth,
            'min_child_weight': min_child_weight,
            'eta': eta,
            'lambda': l2,
            'alpha' : l1,
            'subsample': subsample,
            'colsample_bynode': colsample_bynode,
            'objective':objective,
            'eval_metric':scoring,
        }

        self.max_round_ = max_boost_round
        self.early_stopper_ = xgb.callback.EarlyStopping(rounds = early_stopping_round, min_delta = early_stopping_tol)

    def _get_param_dict(self):
        ret = []
        for md in self.grid_param_['max_depth']:
            for mc in self.grid_param_['min_child_weight']:
                for s in self.grid_param_['subsample']:
                    for cs in self.grid_param_['colsample_bynode']:
                        for et in self.grid_param_['eta']:
                            for l in self.grid_param_['lambda']:
                                for l1 in self.grid_param_['alpha']:
                                    curr = {
                                        'max_depth':md,
                                        'min_child_weight': mc,
                                        'eta': et,
                                        'lambda': l,
                                        'alpha': l1,
                                        'subsample': s,
                                        'colsample_bynode': cs,
                                        'objective':self.grid_param_['objective'],
                                        'eval_metric':self.grid_param_['eval_metric'],
                                    }
                                    ret += [curr]
        return ret

    def _CV(self, sampled_data, seed = 1234):
        x = sampled_data.iloc[:,:-1].values
        y = sampled_data.iloc[:,-1].values
        data_train = xgb.DMatrix(x, label = y)
        best_score = 1e8
        best_param = None
        max_round = None
        allParDicts = self._get_param_dict()
        for params in allParDicts:
            cv_results = xgb.cv(
                params = params,
                dtrain = data_train,
                num_boost_round = self.max_round_,
                seed = seed,
                nfold = 3,
                metrics = params['eval_metric'],
                callbacks = [copy.deepcopy(self.early_stopper_)]
            )
            mean_score = cv_results['test-{}-mean'.format(params['eval_metric'])].min()
            if mean_score < best_score:
                best_score = mean_score
                best_param = params.copy()
                max_round = cv_results['test-{}-mean'.format(params['eval_metric'])].argmin() + 1

        print("    -- final model has {} trees: ".format(max_round), best_param)

        return best_param, max_round

    def _batch_training(self, params, data, n_round):
        last_model = None
        for b in data.batch(split = 'train', size = 3000, shuffle = True):
            curr = b.numpy()
            x = curr[:,:-1]
            y = curr[:,-1]
            data_train = xgb.DMatrix(x, label = y)
            model = xgb.train(
                params = params,
                dtrain = data_train,
                num_boost_round = n_round,
                xgb_model = last_model
            )
            last_model = model

        return model

    def train(self, data, nsample_cv = 6000):

        # first do cv to select best parameters and num_boost_round (CV is always required to get num_boost_round)
        sampled_date = data.sample(n = nsample_cv, split = 'train')
        if data.cache_: #data is dask dataframe, should convert it to pandas
            sampled_date = sampled_date.compute()
        best_pars, n_round = self._CV(sampled_date)

        # train the final model
        if data.cache_: #batch training is needed
            self.model = self._batch_training(best_pars, data, n_round)
        else: #normal training, data is pandas dataframe
            train_data = data.getDatadf(split = 'train')
            x = train_data.iloc[:,:-1].values
            y = train_data.iloc[:,-1].values
            data_train = xgb.DMatrix(x, label = y)
            model = xgb.train(
                params = best_pars,
                dtrain = data_train,
                num_boost_round = n_round,
            )

            self.model = model

    def predict(self, x):
        if not self.model:
            raise ValueError('model has not yet been trained.')
        data = xgb.DMatrix(x)
        return self.model.predict(data)

    def save(self, path):
        if not self.model:
            raise ValueError('model has not yet been trained.')
        joblib.dump(self.model, path)
